/*
 * (C) Copyright IBM Corporation 2015, 2016.
 * (C) Copyright HCL Technologies Limited 2016, 2017.
*/
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

 Author: IBM Corporation and/or HCL Technologies Ltd.
 This module is part of the IBM (R) Rational (R) ClearCase (R)(*)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

 * ClearCase is a trademark of IBM Corporation, registered in many
   jurisdictions, and is used under license.
*/

#include "mvfs_systm.h"         /* The feature macro is defined here. */
#include "mvfs.h"

/* Other MVFS code that uses these routines get the declarations by including
** <lpc/mvfs_lpc_k.h>.
*/
typedef struct mvfs_lpc_handle_pvt {
    mvfs_rcr_hdr_t rcr_hdr;
    MVFS_LPC_MDEP_H_T mh;
} mvfs_lpc_handle_pvt_t;

/* Initialize the LPC system. */
int
mvfs_lpc_initialize(void)
{
    return MVFS_LPC_MDEP_INITIALIZE();
}

/* Teardown the LPC system (undo initialization). */
int
mvfs_lpc_teardown(void)
{
    return MVFS_LPC_MDEP_TEARDOWN();
}

/* Destroy a generic LPC handle (undo handle creation).  Define this first since
** it is used in mvfs_lpc_create below (and it is not externally visible).
*/
static int
mvfs_lpc_destroy(void *h)
{
    int err = 0;
    mvfs_lpc_handle_pvt_t *lh = (mvfs_lpc_handle_pvt_t *)h;

    if ((err = MVFS_LPC_MDEP_DESTROY(lh->mh)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_lpc_destroy: h=%p err=%d, free anyway\n",
                 h, err);
        err = 0;                /* Ignore it, I guess. */
    }
    KMEM_FREE(lh, sizeof(*lh));
    return err;
}

/* Create a generic LPC handle. */
int
mvfs_lpc_create(
    const char *nm,
    CRED_T *cred,
    mvfs_lpc_h *lhp
)
{
    int error = 0;
    mvfs_lpc_handle_pvt_t *lh = NULL;

    *lhp = NULL;

    if ((lh = KMEM_ALLOC(sizeof(*lh), KM_SLEEP)) == NULL) {
        return ENOMEM;
    }

    /* mvfs_rcr initializes with a ref count of 1 so we don't need to hold
     * here.
     */
    mvfs_rcr_init(&lh->rcr_hdr,
                  &mvfs_lpc_destroy,
                  NULL /* Currently unused */);

    if ((error = MVFS_LPC_MDEP_CREATE(nm, cred, &lh->mh)) != 0) {
        KMEM_FREE(lh, sizeof(*lh));
        lh = NULL;
    }

    *lhp = (mvfs_lpc_h)lh;

    return error;
}

/* Flush any pending work associated with the handle. */
int
mvfs_lpc_flush(mvfs_lpc_h lh)
{
    return MVFS_LPC_MDEP_FLUSH(((mvfs_lpc_handle_pvt_t *)lh)->mh);
}

enum clnt_stat
mvfs_lpc_call(
    mvfs_lpc_h lh,
    struct mfs_callinfo *trait,
    int op,
    XID_T xid,
    u_long timeout,
    xdrproc_t xdrargs,
    void *argsp,
    xdrproc_t xdrres,
    void *resp,
    VNODE_T *vw
)
{
    XDR x;
    char *buf = NULL;
    enum clnt_stat error = RPC_SUCCESS;
    timestruc_t start_time;     /* For stats/debug */
    timestruc_t dtime;

    if (lh == NULL ||
        trait == NULL ||
        argsp == NULL ||
        resp == NULL)
    {
        return RPC_SYSTEMERROR;
    }
    if (trait != mfs_viewcall) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_lpc_call: trait=%p, should be %p (&mfs_vwcallstruct)\n",
                 trait, mfs_viewcall);
        return RPC_SYSTEMERROR;
    }

#ifdef MVFS_DEBUG
    /* DEBUG: verify that opnames for trait are up to date */
    if (trait->opnames[op] == NULL) {
        mvfs_log(MFS_LOG_DEBUG, "mvfs_lpc_call: no name for op=%d proto=%lu\n",
                 op, trait->proto);
    }
#endif
    /* Keep stats on calls.  This code is modelled after mvfs_vwcall and
    ** mfscall_int in mvfs_rpcutl.c.
    */
    BUMPSTAT(mfs_clntstat.lpccall); /* Count total calls to LPC. */
    BUMPSTAT(mvfs_rviewopcnt[op]);  /* Count each LPC op. */

    if (vw != NULL) {
        BUMP_VCLNTSTATV(vw, clntstat.lpccall); /* Keep per-view stats. */
    }
    MDKI_HRTIME(&start_time);   /* Get start time for stats. */

    /* XXXODFS: This needs to be replaced by something like a lookaside cache or
     * better yet, some sort of system that can guess what the size of an XDR'd
     * view RPC structure will be. Alternatively, it may be the case that XDR
     * isn't needed at all, in which case, we can just use the arg and res
     * pointers that were passed in.
     */
    if ((buf = KMEM_ALLOC(MVFS_LPC_CALL_MAX_BUFSIZE, KM_SLEEP)) == NULL)
    {
        error = RPC_SYSTEMERROR;
        goto cleanup;
    }

    BZERO(buf, MVFS_LPC_CALL_MAX_BUFSIZE);

    trait->set_xid(argsp, mvfs_get_boottime(), xid);
    MVFS_XDRMEM_CREATE(x, buf, MVFS_LPC_CALL_MAX_BUFSIZE, XDR_ENCODE);

    if (!xdrargs(&x, argsp)) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_lpc_call: Failed to XDR encode arguments");
        error = RPC_CANTENCODEARGS;
        goto cleanup;
    }

    error = MVFS_LPC_MDEP_CALL(((mvfs_lpc_handle_pvt_t *)lh)->mh,
                               op,
                               xid,
                               timeout,
                               buf,
                               xdr_getpos(&x),
                               buf,
                               MVFS_LPC_CALL_MAX_BUFSIZE);

    /* Now see what the call did. */
    if (error != RPC_SUCCESS) {
        mvfs_log(MFS_LOG_INFO,
                 "mvfs_lpc_call: mdep lpc call failed [op = %d, xid = %lx] "
                 "rpcerr = %d\n",
                 op,
                 xid,
                 error);
        goto cleanup;
    }

    /* The call succeeeded, so add to the op histogram. */
    MVFS_BUMPTIME(start_time, dtime, mvfs_rviewoptime[op]);

    if (dtime.tv_sec > mvfs_max_lpcdelay) {
        mvfs_log(MFS_LOG_INFO,
                 "mvfs_lpc_call: View op %s delayed %d seconds!\n",
                 (trait->opnames)[(int)op] ?
                 (trait->opnames)[(int)op] : "unknown operation",
                 (int)dtime.tv_sec);

        /* Update max delay stats per-cpu and per-view. */
        SETSTAT_MAX_DELAY((MVFS_STAT_CNT_T)dtime.tv_sec, lpcmaxdelay);

        if (vw != NULL) {
            SETPVSTAT_MAX_DELAY(vw, (MVFS_STAT_CNT_T)dtime.tv_sec, lpcmaxdelay);
        }
    }
    /* Separate table for cleartext fetch op - expected slowness */
    if (op == VIEW_CLTXT) {
        SET_MAXDELAY(dtime.tv_sec, dtime.tv_nsec, clr, lpc);
    } else {
        SET_MAXDELAY(dtime.tv_sec, dtime.tv_nsec, , lpc);
    }
    MVFS_XDRMEM_CREATE(x, buf, MVFS_LPC_CALL_MAX_BUFSIZE, XDR_DECODE);

    if (!xdrres(&x, resp)) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_lpc_call: Failed to XDR decode response "
                 "[op = %d, xid = %lx]\n",
                 op,
                 xid);
        error = RPC_CANTDECODERES;
        goto cleanup;
    }

    if (trait->get_xid(resp) != xid) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_lpc_call: xid mismatch, sent = %lx, recv = %lx\n",
                 xid,
                 trait->get_xid(resp));
        error = RPC_SYSTEMERROR;
        goto cleanup;
    }

cleanup:
    if (buf != NULL) {
        KMEM_FREE(buf, MVFS_LPC_CALL_MAX_BUFSIZE);
    }
    if (error != RPC_SUCCESS) {
        BUMPSTAT(mfs_clntstat.lpcfail);
    }
    return error;
}

/* Handle the MVFS_CMD_LPC_MESSAGE ioctl call from mvfs_mioctl.c.  The vp we get
** is for the viewtag the caller (rvagent) handles and the message buffer
** pointer is a user space pointer that we need to copyin.  Note, Windows
** doesn't use the MVFS IOCTL for LPC, it calls DeviceControl() directly.
*/
int
mvfs_lpc_ioctl(
    VNODE_T *vp,
    void *datap,
    u_long datalen,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo,
    tbs_status_t *statusp
)
{
    int err = 0;
    mfs_mnode_t *mnp;
    mvfs_lpc_handle_pvt_t *hp;

    ASSERT(MFS_VPISMFS(vp) && MFS_ISAUVIEW(VTOM(vp)));
    mnp = VTOM(vp);

    if ((hp = (mvfs_lpc_handle_pvt_t *)(mnp->mn_view.lpc_handle)) == NULL) {
        err = EINVAL;
        goto done;
    }
    if ((datap == NULL) || (datalen <= 0)) {
        err = EINVAL;
        goto done;
    }
    /* Note, some ioctl calls use statusp to report a tbs_status_t error even if
    ** the call itself was OK (they return error 0).  For instance, if an RPC
    ** was made and returned a response, then the error they return is 0 but
    ** they return the RPC response tbs_status_t in statusp.  In mvfs_mioctl(),
    ** the data->status is initialized to TBS_ST_OK, and then if the ioctl call
    ** returns an error it is set to tbs_errno2status(error) (unless it was
    ** already set to something else).  For us, a "normal" error is ESHUTDOWN,
    ** which we return when the view (and its lpc_handle) are being removed (by
    ** a rmviewtag ioctl).  In that case we convert it to TBS_ST_ABORT and make
    ** this ioctl return OK.
    */
    err = MVFS_LPC_MDEP_IOCTL(hp->mh,
                              datap,
                              datalen,
                              cd,
                              callinfo);
  done:
    if (err == ESHUTDOWN) {
        *statusp = TBS_ST_ABORT;
        err = 0;
    }
    return err;
}
static const char vnode_verid_mvfs_lpc_c[] = "$Id:  a684bcf2.bf2911e7.8732.44:37:e6:71:22:3b $";
